from functools import reduce
from itertools import chain

from django.core.urlresolvers import resolve, reverse
from django.core.exceptions import ObjectDoesNotExist
from django.db.models import OneToOneField
from django.forms.fields import DateTimeField
from django.forms.models import modelform_factory, inlineformset_factory
from django.http import Http404
from django.shortcuts import redirect
from django.utils.functional import cached_property
from django.views.generic.edit import FormMixin

from ..forms import FrozenSelect, DateTimePicker
from ..components import Form
from ..utils.functools import just, call, swap, throw, \
    pluck, first, select, separate, emptyless, get, omit, pick, map_keys, map_values
from ..utils.translation import _, inflect, get_string, get_model_names
from .base import BaseView


class ModelFormMixin:
    @cached_property
    def model(self):
        return getattr(
            self,
            'model_class',
            hasattr(self.form_class, '_meta') \
            and self.form_class._meta.model or None)

    @cached_property
    def instance(self):
        return call(getattr(self, 'get_instance', lambda: None))

    def get_model_name(self, key='single', fn=get_string, model=None):
        return get(key, dict(
            zip(('single', 'plural'),
            map(fn, get_model_names(model or self.model)))))

    def get_model_fields(self, model=None):
        return pluck((model or self.model)._meta.fields[1:], 'name')

    def get_fields(self):
        return self.form_class and self.form_class._meta.fields \
            or getattr(self, 'fields', self.get_model_fields())

    def get_choices(self, model, field):
        return list(model._meta.get_field(field).choices)

    def get_relation_fields(self):
        return [] if not self.model else filter(
            lambda field: isinstance(field, OneToOneField),
            emptyless(map(
                lambda field: field in self.model._meta.get_all_field_names() \
                    and self.model._meta.get_field(field) or None,
                self.get_fields())))

    def get_form_class(self):
        return self.form_class or modelform_factory(
            self.model, fields=self.get_fields())

    def get_form_data(self, **kw):
        result = pick(kw, 'prefix', 'initial', 'instance')
        if self.request.POST:
            result['data'] = self.request.POST
        if self.request.FILES:
            result['files'] = self.request.FILES
        return result

    def get_form(self):
        form = self.get_form_class()(**self.get_form_kwargs())
        for field in pluck(self.get_relation_fields(), 'name'):
            form.fields.pop(field)
        return self.prepare_form(form)

    def get_relation(self, field, **kw):
        kw = kw or {'fields': self.get_model_fields(field.rel.to)}
        form = modelform_factory(field.rel.to, **kw)(
            **self.get_form_data(
                prefix=field.name,
                instance=getattr(self.instance, field.name, None)))
        return {'field': field, 'form': form}

    def get_relations(self):
        return emptyless(map(
            lambda field: swap(field, getattr(
                self,
                'get_{0}_relation'.format(field.name),
                self.get_relation)),
            self.get_relation_fields()))

    def prepare_form(self, form):
        # TODO: DateField, settings
        overrides = dict(
            {DateTimeField: [DateTimePicker, {}]},
            **getattr(self, 'form_display', {}))

        for field in form:
            widget = overrides.get(field.field.__class__, None)
            if widget:
                field.field.widget = widget[0](**widget[1])
        return form


class FormsetsMixin(ModelFormMixin):
    def get_formsets(self):
        return []

    def fill_matrix(self, factory, fields):
        choices = dict(map(
            lambda field: (field, list(filter(
                lambda choice: not choice in list(map(
                    lambda instance: getattr(instance, field),
                    factory.model.objects.filter(
                        **{factory.fk.name: self.instance}))),
                map(first, self.get_choices(factory.model, field))))),
            fields))
        return list(map(
            lambda i: map_values(lambda c: select(i, c), choices),
            range(0, factory.extra)))

    def get_formset_factory(self, formset_model, **kw):
        if 'matrix' in kw:
            max_num = max(map(
                lambda field: len(self.get_choices(formset_model, field)),
                kw['matrix']))
            kw = dict({k: max_num for k in ('extra', 'max_num')}, **kw)
            kw['widgets'] = {f: FrozenSelect() for f in kw.pop('matrix', [])}
        if not 'fields' in kw and not 'form' in kw:
            kw['fields'] = self.get_model_fields(formset_model)
        return inlineformset_factory(
            kw.pop('model', self.model), formset_model, **kw)

    def get_formset_data(self, factory, **kw):
        result = self.get_form_data(instance=self.instance, **kw)
        if 'matrix' in kw:
            result['initial'] = self.fill_matrix(factory, kw['matrix'])
        return result

    def inline_formset(self, formset_model, **kw):
        factory = self.get_formset_factory(formset_model, **kw.copy())
        prefix = self.get_model_name('plural', model=factory.model)
        get_data = getattr(self, 'get_{0}_formset_data'.format(prefix), self.get_formset_data)
        formset = factory(**get_data(factory, **dict(kw, prefix=prefix)))

        if formset.can_delete and formset.validate_min:
            for form in formset.forms[:formset.min_num]:
                form.fields['DELETE'].widget.attrs['disabled'] = True
        return formset


class FormView(BaseView, FormsetsMixin, FormMixin):
    use_prefix = False
    components = {'content': Form}
    form_props = {}
    success_redirect = 'index'
    success_message = None
    actions = [{
        'icon': 'check-square',
        'title': _('Save'),
        'level': 'success'}]

    @property
    def url_match(self):
        return resolve(self.request.path)

    def get_attrs_dict(self, *attrs):
        return dict(map(lambda x: (x, call(getattr(self, 'get_' + x))), attrs))

    def get_prefix(self):
        return self.prefix or (self.use_prefix and self.about or None)

    def get_form_kwargs(self):
        return self.get_form_data(**dict(
            {} if not self.instance else {'instance': self.instance},
            **self.get_attrs_dict('prefix', 'initial')))

    def get_members(self):
        result = self.get_attrs_dict('form', 'relations', 'formsets')
        return result

    def get_all_forms(self, **kw):
        return chain(
            [kw['form']],
            pluck(kw['relations'], 'form'),
            *pluck(kw['formsets'], 'forms'))

    def get_media(self, **kw):
        return reduce(
            lambda x, y: x + y,
            pluck(self.get_all_forms(**kw), 'media'))

    def get_actions(self):
        return self.actions

    def get_form_props(self):
        return dict(self.form_props, actions=self.get_actions())

    def get_context_data(self, **kw):
        members = kw or self.get_members()
        return dict(
            super().get_context_data(**members),
            **map_keys(
                lambda key: key if not self.use_prefix else '_'.join(
                    [self.get_prefix(), key]),
                dict(
                    members,
                    props=self.get_form_props())))

    def get_success_message(self, **kw):
        return self.success_message

    def get_success_redirect(self, **kw):
        return reverse(getattr(self, 'success_redirect', self.url_match.url_name))

    def success_finish(self, **kw):
        self.send_message('success', self.get_success_message(**kw))
        return redirect(self.get_success_redirect(**kw))

    def save_relations(self, form, *relations):
        for rel in relations:
            if rel['form'].has_changed():
                setattr(form.instance, rel['field'].name, rel['form'].save())

    def save_form(self, form):
        hasattr(form, 'instance') and form.instance.save()

    def save_formsets(self, form, *formsets):
        for formset in formsets:
            formset.instance = form.instance
            formset.save()

    def process_valid(self, **kw):
        self.save_relations(kw['form'], *kw['relations'])
        self.save_form(kw['form'])
        self.save_formsets(kw['form'], *kw['formsets'])
        return self.success_finish(**kw)

    def process_invalid(self, **kw):
        errors = chain(*map(call, pluck(
            self.get_all_forms(**kw), 'non_field_errors')))
        for message in errors:
            self.send_message('error', message)
        return self.render_to_response(self.get_context_data(**kw))

    def check_is_valid(self, **kw):
        return all(map(lambda x: x.is_valid(), self.get_all_forms(**kw)))

    def process(self, **kw):
        if self.check_is_valid(**kw):
            map(
                lambda x: x.save(commit=False),
                chain(
                    hasattr(kw['form'], 'save') and [kw['form']] or [],
                    *pluck(kw['relations'], 'form')))
            return self.process_valid(**kw)
        return self.process_invalid(**kw)

    def post(self, request, *args, **kw):
        return self.process(**self.get_members())


class ObjectView(FormView):
    def get_instance(self):
        pk = self.url_match.kwargs.get('pk')
        try:
            return pk and self.model.objects.get(pk=pk)
        except ObjectDoesNotExist:
            raise Http404

    def get_form_props(self):
        return dict(
            super().get_form_props(),
            panels=True,
            title=self.instance and str(self.instance) or self.model._meta.verbose_name)

    def get_success_redirect(self, **kw):
        return reverse(self.get_model_name('plural'))

    def get_success_message(self, **kw):
        return inflect(_(
            'The %(name)s "%(obj)s" was changed successfully.' \
            if self.url_match.kwargs else \
            'The %(name)s "%(obj)s" was added successfully.') % {
                'name': self.get_model_name(fn=just),
                'obj': str(kw['form'].instance)})



from django import forms
from django.templatetags.static import static
from .utils.functools import first, avg


class EmptyWidget(forms.widgets.Widget):
    def render(self, *args, **kw):
        return ''


class Slider(forms.widgets.TextInput):
    class Media:
        css = {'all': [static('seiyria-bootstrap-slider/dist/css/bootstrap-slider.min.css')]}
        js = [
            static('seiyria-bootstrap-slider/dist/bootstrap-slider.min.js'),
            static('widgets/slider.js')]

    def __init__(self, attrs=None):
        attrs = attrs or {}
        if not 'range' in attrs:
            attrs['range'] = range(0, 100, 1)
        attrs.update(dict(zip(('min', 'max', 'step'), map(
            lambda a: getattr(attrs['range'], a),
            ('start', 'stop', 'step')))))
        attrs['value'] = attrs.pop('value', int(
            avg(attrs['range'].start, attrs.pop('range').stop)))
        super().__init__(dict(map(
            lambda t: ('data-slider-' + t[0], t[1]),
            attrs.items())))


class FrozenSelect(forms.widgets.Select):
    def render(self, name, value, attrs=None, choices=()):
        return ''.join([
            str(list(filter(lambda c: c[0] == value, self.choices))[0][1]),
            forms.widgets.HiddenInput().render(name, value)])


class DateTimePicker(forms.DateTimeInput):
    # language: lang = translation.get_language()
    # formats from settings
    # json options
    # fa icons

    class Media:
        css = {'all': [static(
            'eonasdan-bootstrap-datetimepicker/build/css/bootstrap-datetimepicker.min.css')]}
        js = [static(
            'eonasdan-bootstrap-datetimepicker/build/js/bootstrap-datetimepicker.min.js')]

    def render(self, name, value, attrs=None):
        return '{input}<script>$("#{id}").parent().datetimepicker({options})</script>'.format(
            input=super().render(name, value, attrs),
            id=attrs['id'],
            options='')



from importlib import import_module
from itertools import chain
import re

from classytags.arguments import MultiKeywordArgument
from classytags.core import Options
from classytags.core import Tag as ClassyTag
from django.template import Library, TemplateSyntaxError

from ..utils.functools import call, unpack_args, reduce_dict, map_values, \
    truthful, emptyless, pick, omit
from ..utils.translation import normalize


register = Library()
register.era = lambda cls: register.tag(normalize(cls.__name__), cls)


@register.era
class Import(ClassyTag):
    def __init__(self, parser, tokens):
        for module in tokens.contents.split(' ')[1:]:
            path = '.'.join(chain(
                [module.split('.')[0]],
                ['components'],
                [] if not '.' in module else module.split('.')[1:]))
            try:
                parser.tags.update(import_module(path).register.tags)
            except ImportError:
                raise TemplateSyntaxError('can not import {0}'.format(path))


class HTMLTag:
    def __init__(self, el):
        self.el = str(el)

    def __str__(self):
        return self.el

    def modify_tag(self, diff):
        tag, *rest = self.el.split('>')
        self.el = '>'.join([diff(tag)] + rest)

    def modify_content(self, content, fmt):
        self.el = re.sub(r'>(.*)<', fmt.format(content), self.el)

    def append(self, content):
        self.modify_content(content, r'>\g<1>{0}<')

    def prepend(self, content):
        self.modify_content(content, r'>{0}\g<1><')

    def add_attr(self, name, val=None):
        self.modify_tag(lambda tag: ' '.join([
            tag.endswith('/') and tag.rstrip('/') or tag,
            val is None and name or '{0}="{1}"'.format(name, val),
            tag.endswith('/') and '/' or '']))

    def add_class(self, *cls):
        if not any(cls):
            return
        if not self.has_attr('class'):
            self.add_attr('class', ' '.join(cls))
        else:
            self.modify_tag(
                lambda tag: re.sub(
                    r'class="([^"]*)"',
                    r'class="\g<1> {0}"'.format(' '.join(cls)),
                    tag))

    def has_attr(self, name):
        return name in self.el.split('>')[0]

    def empty(self):
        self.el = ''


class Props(dict):
    def __getattr__(self, name):
        if name in self:
            return self.get(name)
        raise AttributeError('Missing prop: ' + name)


class Component(ClassyTag):
    def __init__(self, parser=None, tokens=None):
        self.blocks = {}
        self.set_options()
        parser and super().__init__(parser, tokens)

    @classmethod
    def as_string(cls, request=None, **kw):
        return cls().render_tag({'request': request}, **kw)

    def inject(self, cls, props=None, nodelist=None):
        return '' if not cls else cls().render_tag(self.context, **dict(
            props or {}, **(nodelist is not None and {'nodelist': nodelist} or {})))

    def join(self, *components):
        return ''.join(map(lambda x: self.inject(x), components))

    def build(self, args, prefix='render'):
        return map(lambda arg: call(getattr(self, '_'.join([prefix, arg]))), args)

    def get_defaults(self):
        return {}

    def get_class_set(self, *args, prefix='', include=''):
        return ' '.join(emptyless(
            chain([include], map(
                lambda k: '-'.join(emptyless([prefix, k])),
                truthful(pick(self.props, *args)).keys()))))

    def resolve_props(self):
        return {}

    def set_options(self, **kw):
        self.options = Options(MultiKeywordArgument('mka', required=False), **kw)

    def set_props(self, **kw):
        self.props = Props(
            self.get_defaults(),
            **dict(kw, **map_values(lambda val: val.render(self.context), self.blocks)))
        self.props.update(self.resolve_props())

    def render_tag(self, context, mka=None, **kw):
        self.context = context
        self.request = context['request']
        self.set_props(**dict(mka or {}, **kw))
        return self.render_dom()

    def render_dom(self):
        self.dom = HTMLTag(self.DOM())
        self.tweak()
        return str(self.dom)

    def DOM(self):
        raise NotImplementedError

    def tweak(self):
        if not 'attrs' in self.props and 'class' in self.props:
            self.dom.add_class(self.props['class'])


@register.era
class Inject(Component):
    def DOM(self):
        return '' if not self.props.component else self.inject(
            lambda: self.props.component,
            omit(self.props, 'component'))


class ComplexComponent(Component):
    parts = []

    def get_defaults(self):
        return {'nodelist': ''}

    def set_options(self, **kw):
        return super().set_options(**(kw or {'blocks': chain(
            [] if not self.parts else [(self.parts[0], 'nodelist')],
            [] if not self.parts else list(map(unpack_args(
                lambda i, b: (b, self.parts[i - 1])),
                enumerate(self.parts[1:], start=1))),
            [(
                'end-' + normalize(self.__class__.__name__),
                not self.parts and 'nodelist' or self.parts[-1])])}))


@register.era
class Tag(ComplexComponent):
    el = 'div'
    nobody = False
    inline = False

    def get_nodelist(self):
        return self.props.nodelist

    def resolve_attrs(self):
        return {}

    def resolve_tag(self):
        attrs = dict(
            self.props.get('attrs', {}),
            **self.resolve_attrs())
        if 'id' in self.props:
            attrs['id'] = self.props.id
        if 'class' in self.props:
            attrs['class'] = ' '.join(emptyless([
                attrs.get('class', ''),
                self.props['class']]))
        return {
            'el': self.props.get('el', getattr(self, 'el')),
            'attrs': ' '.join(reduce_dict(
                lambda k, v: v and '{0}="{1}"'.format(k, v) or k,
                attrs))}

    def set_options(self):
        return super().set_options(**(
            (self.inline or self.nobody) and {'blocks': []} or {}))
    
    def set_props(self, **kw):
        super().set_props(**kw)
        self.props.update(self.resolve_tag())
        not self.nobody and self.props.update({'nodelist': self.get_nodelist()})

    def DOM(self):
        return ''.join([
            '<{el} {attrs}',
            '/>' if self.nobody else '>{nodelist}</{el}>']) \
        .format(**self.props)


@register.era
class Content(Component):
    def DOM(self):
        return self.inject(self.context['components'].get('content', None))



from itertools import chain
from urllib.parse import urlencode

from django.conf import settings
from django.db.models import Count
from django.db.models.fields import BooleanField, FieldDoesNotExist
from django.db.models.fields.related import ManyToOneRel
from django.shortcuts import redirect
from django.utils.functional import cached_property
from django.views.generic.list import BaseListView

from ..components import ObjectsList, ChangeList
from ..utils.functools import just, call, first, pluck, \
    map_keys, map_values, reduce_dict, filter_dict
from ..utils.translation import _, get_string, get_model_names, verbose_choices
from .base import BaseView


class ListView(BaseView, BaseListView):
    components = {'content': ObjectsList}
    list_display = []

    def lookup_field(self, obj, field, next_obj=getattr, end_obj=getattr):
        if not '__' in field:
            return end_obj(obj, field)
        return self.lookup_field(
            next_obj(obj, field.split('__')[0]),
            '__'.join(field.split('__')[1:]),
            next_obj=next_obj,
            end_obj=end_obj)

    def get_model_field(self, field):
        try:
            return self.lookup_field(
                self.model,
                get_string(field),
                next_obj=lambda model, field: model._meta.get_field(field).rel.to,
                end_obj=lambda model, field: model._meta.get_field(field))
        except FieldDoesNotExist:
            return {'name': field}

    def get_list_view(self, op, value=None):
        return value or getattr(self, 'list_' + op)

    def lookup_field_display(self, obj, field, model=True):
        method = 'get_{0}_display'.format(field)
        if isinstance(obj, self.model) and hasattr(self, method):
            return getattr(self, method)(obj)
        elif model:
            return call(getattr(obj, method, lambda: getattr(obj, field)))
        raise AttributeError('Missing display data for "{0}"'.format(field))

    def display_column(self, field):
        if isinstance(field, dict):
            return field['name']
        elif isinstance(field, ManyToOneRel):
            return field.field.model._meta.verbose_name_plural
        else:
            return field.verbose_name

    def display_field(self, name, field, obj):
        if isinstance(field, dict):
            return self.lookup_field_display(
                obj,
                get_string(field['name']).replace(' ', '_'),
                model=False)
        elif '__' in get_string(name):
            return self.lookup_field(obj, name, end_obj=self.lookup_field_display)
        else:
            value = self.lookup_field_display(obj, field.name)
            if isinstance(field, ManyToOneRel) and not isinstance(value, str):
                return value.count()
            return value

    def get_thead_items(self):
        return list(map(
            self.display_column, map(
                self.get_model_field,
                self.get_list_view('display'))))

    def get_tbody_items(self, objects):
        return list(map(
            lambda obj: {
                'pk': obj.pk,
                'fields': list(map(
                    lambda name: self.display_field(
                        name, self.get_model_field(name), obj),
                    self.get_list_view('display')))},
            objects))

    def get_context_data(self, **kw):
        data = super().get_context_data(**kw)
        return dict(
            data,
            thead=self.get_thead_items(),
            tbody=self.get_tbody_items(data['object_list']))


class AdminView(ListView):
    components = {'content': ChangeList}
    list_filter = []
    list_counters = []
    list_sort = []
    actions = [{
        'icon': 'plus-square',
        'title': _('Add'),
        'level': 'success',
        'link': {'rel': 'add'}}]

    @property
    def objects(self):
        if not hasattr(self, '_objects'):
            self._objects = self.get_queryset(ignore_state=True)
        return self._objects

    @property
    def filters(self):
        return self.get_filters()

    @property
    def filter_keys(self):
        return map(
            first, map(
                lambda name: self.get_filter(name, self.get_model_field(name)),
                self.get_list_view('filter')))

    @property
    def sort_keys(self):
        return map(
            lambda name: name.replace('__', '.'),
            self.get_list_view('sort'))

    @property
    def active_filters(self):
        return list(map(
            first, filter(
                lambda f: f[1] in self.request.GET,
                zip(
                    self.list_filter,
                    map(
                        lambda key: '-'.join(['filter', key]),
                        self.filter_keys)))))

    def get_state(self, method):
        return map_keys(
            lambda k: k.replace('.', '__'),
            map_values(
                lambda v: v if not v in ('True', 'False') else v[0] == 'T',
                filter_dict(
                    lambda k, v: k in list(getattr(self, '_'.join([method, 'keys']))),
                    map_keys(
                        lambda k: k.split('-')[1],
                        filter_dict(
                            lambda k, v: k.startswith(method),
                            self.request.GET)))))

    def get_queryset(self, ignore_state=False):
        qs = self.model.objects.all() if self.queryset is None else self.queryset
        if not ignore_state:
            states = {k: self.get_state(k) for k in ('filter', 'sort')}
            if states['filter']:
                qs = qs.filter(**states['filter'])
            if states['sort']:
                qs = qs.order_by(*reduce_dict(
                    lambda k, v: ('-' if v is False else '') + k,
                    states['sort']))
        return qs

    def get_list_view(self, op, value=None):
        if op == 'display':
            value = list(filter(
                lambda f: not f in self.active_filters,
                (value or self.list_display)))
        return super().get_list_view(op, value)

    def get_actions(self):
        return self.actions

    def get_generic_filter(self, name, choices):
        result = []
        name = name.replace('__', '.')
        values = pluck(self.objects, name)
        for choice in choices:
            count = values.count(choice[0])
            if count:
                result.append(list(chain(choice, [count])))
        return [name, result]

    def get_filter(self, name, field):
        storage = '_{0}_filter'.format(name)
        if not hasattr(self, storage):
            if isinstance(field, BooleanField):
                result = self.get_generic_filter(
                    name, [(True, _('Yes')), (False, _('No'))])
            elif field.is_relation:
                result = self.get_generic_filter(
                    name + '__pk', map(
                        lambda obj: (obj.pk, str(obj)),
                        field.rel.to.objects.all()))
            elif hasattr(field, 'choices'):
                result = self.get_generic_filter(name, field.choices)
            else:
                method = 'get_{0}_filter'.format(get_string(name))
                result = getattr(self, method)(name, field)
            setattr(self, storage, result)
        return getattr(self, storage)

    def get_filters(self):
        return list(map(
            lambda name: chain(
                [self.display_column(self.get_model_field(name))],
                self.get_filter(name, self.get_model_field(name)),
                [name in self.get_list_view('counters')]),
            self.get_list_view('filter')))

    def get_thead_items(self):
        return map(
            lambda name: [
                self.display_column(self.get_model_field(name)),
                name in self.get_list_view('sort') and name.replace('__', '.')],
            self.get_list_view('display'))

    def get_context_data(self, **kw):
        return dict(
            super().get_context_data(**kw),
            actions=self.get_actions(),
            filters=self.filters)

    def get_redirect_features(self):
        if not self.objects:
            return self.navigate('-'.join([
                get_string(get_model_names(self.model)[0]),
                'add']))
        if self.list_filter and not self.request.GET:
            kw = {}
            for (title, key, choices, counters) in self.filters:
                choices = list(map(first, choices))
                if len(choices) == 1:
                    kw['-'.join(['filter', key])] = choices[0]
            if kw:
                return redirect('?'.join([self.request.path, urlencode(kw)]))

    def get(self, request, *args, **kwargs):
        return self.get_redirect_features() \
            or super().get(request, *args, **kwargs)
